@@ -289,11 +289,11 @@ bool parseAtom(const char*&scursor, const char *const send,
 // FIXME: Remove this and the other parseToken() method. add a new one where "result" is a
 //        QByteArray.
 bool parseToken(const char*&scursor, const char *const send,
-                QString &result, bool allow8Bit)
+                QString &result, ParseTokenFlags flags)
 {
     QPair<const char *, int> maybeResult;
 
-    if (parseToken(scursor, send, maybeResult, allow8Bit)) {
+    if (parseToken(scursor, send, maybeResult, flags)) {
         result = QString::fromLatin1(maybeResult.first, maybeResult.second);
         return true;
     }
@@ -302,7 +302,7 @@ bool parseToken(const char*&scursor, const char *const send,
 }
 
 bool parseToken(const char*&scursor, const char *const send,
-                QPair<const char *, int> &result, bool allow8Bit)
+                QPair<const char *, int> &result, ParseTokenFlags flags)
 {
     bool success = false;
     const char *start = scursor;
@@ -312,10 +312,12 @@ bool parseToken(const char*&scursor, const char *const send,
         if (ch > 0 && isTText(ch)) {
             // TText: OK
             success = true;
-        } else if (allow8Bit && ch < 0) {
+        } else if ((flags & ParseTokenAllow8Bit) && ch < 0) {
             // 8bit char: not OK, but be tolerant.
             KMIME_WARN_8BIT(ch);
             success = true;
+        } else if ((flags & ParseTokenRelaxedTText) && ch == '/') {
+            success = true;
         } else {
             // CTL or tspecial - marking the end of the atom:
             // re-set sursor to point to the offending
@@ -346,7 +348,6 @@ bool parseGenericQuotedString(const char *&scursor, const char *const send,
                               QString &result, bool isCRLF,
                               const char openChar, const char closeChar)
 {
-    char ch;
     // We are in a quoted-string or domain-literal or comment and the
     // cursor points to the first char after the openChar.
     // We will apply unfolding and quoted-pair removal.
@@ -356,7 +357,7 @@ bool parseGenericQuotedString(const char *&scursor, const char *const send,
     assert(*(scursor - 1) == openChar || *(scursor - 1) == closeChar);
 
     while (scursor != send) {
-        ch = *scursor++;
+        char ch = *scursor++;
 
         if (ch == closeChar || ch == openChar) {
             // end of quoted-string or another opening char:
@@ -451,9 +452,7 @@ bool parseGenericQuotedString(const char *&scursor, const char *const send,
                 scursor = oldscursor;
             }
             // fall through
-#if QT_VERSION >= QT_VERSION_CHECK(5,8,0)
-        Q_FALLTHROUGH();
-#endif
+            Q_FALLTHROUGH();
         }
         default:
             KMIME_WARN_IF_8BIT(ch);
@@ -637,10 +636,8 @@ bool parsePhrase(const char *&scursor, const char *const send,
                 // parse as atom:
                 scursor = oldscursor;
             }
-#if QT_VERSION >= QT_VERSION_CHECK(5,8,0)
-        Q_FALLTHROUGH();
-#endif
-        // fall though...
+            Q_FALLTHROUGH();
+            // fall though...
 
         default: //atom
             tmp.clear();
@@ -1233,7 +1230,7 @@ bool parseParameter(const char *&scursor, const char *const send,
     //
     // FIXME: maybeAttribute should be a QByteArray
     QString maybeAttribute;
-    if (!parseToken(scursor, send, maybeAttribute, false /* no 8bit */)) {
+    if (!parseToken(scursor, send, maybeAttribute, ParseTokenNoFlag)) {
         return false;
     }
 
@@ -1281,7 +1278,7 @@ bool parseParameter(const char *&scursor, const char *const send,
         }
     } else {
         // value is a token:
-        if (!parseToken(scursor, send, maybeValue.qpair, false /* no 8bit */)) {
+        if (!parseToken(scursor, send, maybeValue.qpair, ParseTokenRelaxedTText)) {
             scursor = oldscursor;
             result = qMakePair(maybeAttribute.toLower(), QStringOrQPair());
             return false; // this case needs further processing by upper layers!!
@@ -1714,7 +1711,7 @@ static bool parseAlphaNumericTimeZone(const char *&scursor,
     }
 
     QPair<const char *, int> maybeTimeZone(nullptr, 0);
-    if (!parseToken(scursor, send, maybeTimeZone, false /*no 8bit*/)) {
+    if (!parseToken(scursor, send, maybeTimeZone, ParseTokenNoFlag)) {
         return false;
     }
     for (int i = 0 ; i < timeZonesLen ; ++i) {
@@ -2026,40 +2023,53 @@ bool parseDateTime(const char *&scursor, const char *const send,
     return true;
 }
 
+namespace {
+
+Headers::Base *extractHeader(const QByteArray &head, const int headerStart, int &endOfFieldBody)
+{
+    Headers::Base *header = {};
+
+    int startOfFieldBody = head.indexOf(':', headerStart);
+    if (startOfFieldBody < 0) {
+        return nullptr;
+    }
+
+    const char *rawType = head.constData() + headerStart;
+    const size_t rawTypeLen = startOfFieldBody - headerStart;
+
+    startOfFieldBody++; //skip the ':'
+    if (startOfFieldBody < head.size() - 1 &&  head[startOfFieldBody] == ' ') { // skip the space after the ':', if there's any
+        startOfFieldBody++;
+    }
+
+    bool folded = false;
+    endOfFieldBody = findHeaderLineEnd(head, startOfFieldBody, &folded);
+
+    // We might get an invalid mail without a field name, don't crash on that.
+    if (rawTypeLen > 0) {
+        header = HeaderFactory::createHeader(rawType, rawTypeLen);
+    }
+    if (!header) {
+        //qWarning() << "Returning Generic header of type" << rawType;
+        header = new Headers::Generic(rawType, rawTypeLen);
+    }
+    if (folded) {
+        const auto unfoldedBody = unfoldHeader(head.constData() + startOfFieldBody, endOfFieldBody - startOfFieldBody);
+        header->from7BitString(unfoldedBody);
+    } else {
+        header->from7BitString(head.constData() + startOfFieldBody, endOfFieldBody - startOfFieldBody);
+    }
+
+    return header;
+}
+
+}
+
 Headers::Base *extractFirstHeader(QByteArray &head)
 {
     int endOfFieldBody = 0;
-    bool folded = false;
-    Headers::Base *header = nullptr;
-
-    int startOfFieldBody = head.indexOf(':');
-
-    if (startOfFieldBody > -1) {      //there is another header
-        // Split the original data
-        head[startOfFieldBody] = '\0';
-        // rawType references the actual data from 'head'
-        QByteArray rawType = QByteArray::fromRawData(head.constData(), startOfFieldBody);
-
-        startOfFieldBody++; //skip the ':'
-        if (head[startOfFieldBody] == ' ') {   // skip the space after the ':', if there
-            startOfFieldBody++;
-        }
-        endOfFieldBody = findHeaderLineEnd(head, startOfFieldBody, &folded);
-        // rawFieldBody references actual data from 'heaed'
-        QByteArray rawFieldBody = QByteArray::fromRawData(head.constData() + startOfFieldBody, endOfFieldBody - startOfFieldBody);
-        if (folded) {
-            rawFieldBody = unfoldHeader(rawFieldBody);
-        }
-        // We might get an invalid mail without a field name, don't crash on that.
-        if (!rawType.isEmpty()) {
-            header = HeaderFactory::createHeader(rawType);
-        }
-        if (!header) {
-            //qWarning() << "Returning Generic header of type" << rawType;
-            header = new Headers::Generic(rawType.constData());
-        }
-        header->from7BitString(rawFieldBody);
-
+    auto header = extractHeader(head, 0, endOfFieldBody);
+    if (header) {
         head.remove(0, endOfFieldBody + 1);
     } else {
         head.clear();
@@ -2094,11 +2104,17 @@ void extractHeaderAndBody(const QByteArray &content, QByteArray &header, QByteAr
 QVector<Headers::Base*> parseHeaders(const QByteArray &head)
 {
     QVector<Headers::Base*> ret;
-    Headers::Base *h;
 
-    QByteArray copy = head;
-    while ((h = extractFirstHeader(copy))) {
-        ret << h;
+    int cursor = 0;
+    while (cursor < head.size()) {
+        const int headerStart = cursor;
+        int endOfFieldBody;
+        if (auto header = extractHeader(head, headerStart, endOfFieldBody)) {
+            ret << header;
+            cursor = endOfFieldBody + 1;
+        } else {
+            break;
+        }
     }
 
     return ret;
